@@ -15,10 +15,14 @@ logger = logging.getLogger('jply')
 logger.setLevel(logging.INFO)
 
 ch = logging.StreamHandler(sys.stdout)
-ch.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s — %(message)s"))
+ch.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(message)s"))
 logger.addHandler(ch)
 logger.propagate = False
 
+IFAS_BASE_URL = "https://ifas.prod-row.jlrmotor.com/ifas/jlr"
+IFOP_BASE_ULR = "https://ifop.prod-row.jlrmotor.com/ifop/jlr"
+IF9_BASE_URL = "https://if9.prod-row.jlrmotor.com/if9/jlr"
+
 
 class Connection(object):
     """Connection to the JLR Remote Car API"""
@@ -26,10 +30,13 @@ class Connection(object):
     def __init__(self,
                  email='',
                  password='',
-                 device_id='', ):
+                 device_id='',
+                 refresh_token=''):
         """Init the connection object
 
         The email address and password associated with your Jaguar InControl account is required.
+        A device Id can optionally be specified. If not one will be generated at runtime.
+        A refresh token can be supplied for authentication instead of a password
         """
         self.email = email
 
@@ -38,10 +45,16 @@ class Connection(object):
         else:
             self.device_id = str(uuid.uuid4())
 
-        self.oauth = {
-            "grant_type": "password",
-            "username": email,
-            "password": password}
+        if refresh_token:
+            self.oauth = {
+                "grant_type": "refresh_token",
+                "refresh_token": refresh_token}
+        else:
+            self.oauth = {
+                "grant_type": "password",
+                "username": email,
+                "password": password}
+
         self.expiration = 0  # force credential refresh
 
         self.connect()
@@ -70,12 +83,15 @@ class Connection(object):
         logger.info("Connecting...")
         auth = self.__authenticate(data=self.oauth)
         self.__register_auth(auth)
-        logger.info("1/3 authenticated")
         self.__set_header(auth['access_token'])
+        logger.info("[+] authenticated")
+        self.__register_device_and_log_in()
+
+    def __register_device_and_log_in(self):
         self.__register_device(self.head)
-        logger.info("2/3 device id registered")
+        logger.info("1/2 device id registered")
         self.__login_user(self.head)
-        logger.info("3/3 user logged in, user id retrieved")
+        logger.info("2/2 user logged in, user id retrieved")
 
     def __open(self, url, headers=None, data=None):
         req = Request(url, headers=headers)
@@ -103,11 +119,12 @@ class Connection(object):
         self.head = {
             "Authorization": "Bearer %s" % access_token,
             "X-Device-Id": self.device_id,
+            "x-telematicsprogramtype", "jlrpy",
             "Content-Type": "application/json"}
 
     def __authenticate(self, data=None):
         """Raw urlopen command to the auth url"""
-        url = "https://jlp-ifas.wirelesscar.net/ifas/jlr/tokens"
+        url = "%s/tokens" % IFAS_BASE_URL
         auth_headers = {
             "Authorization": "Basic YXM6YXNwYXNz",
             "Content-Type": "application/json",
@@ -117,7 +134,7 @@ class Connection(object):
 
     def __register_device(self, headers=None):
         """Register the device Id"""
-        url = "https://jlp-ifop.wirelesscar.net/ifop/jlr/users/%s/clients" % self.email
+        url = "%s/users/%s/clients" % (IFOP_BASE_ULR, self.email)
         data = {
             "access_token": self.access_token,
             "authorization_token": self.auth_token,
@@ -129,7 +146,7 @@ class Connection(object):
 
     def __login_user(self, headers=None):
         """Login the user"""
-        url = "https://jlp-ifoa.wirelesscar.net/if9/jlr/users?loginName=%s" % self.email
+        url = "%s/users?loginName=%s" % (IF9_BASE_URL, self.email)
         user_login_header = headers.copy()
         user_login_header["Accept"] = "application/vnd.wirelesscar.ngtp.if9.User-v3+json"
 
@@ -137,25 +154,37 @@ class Connection(object):
         self.user_id = user_data['userId']
         return user_data
 
+    def refresh_tokens(self):
+        """Refresh tokens."""
+        self.oauth = {
+            "grant_type": "refresh_token",
+            "refresh_token": self.refresh_token}
+
+        auth = self.__authenticate(self.oauth)
+        self.__register_auth(auth)
+        self.__set_header(auth['access_token'])
+        logger.info("[+] Tokens refreshed")
+        self.__register_device_and_log_in()
+
     def get_vehicles(self, headers):
         """Get vehicles for user"""
-        url = "https://jlp-ifoa.wirelesscar.net/if9/jlr/users/%s/vehicles?primaryOnly=true" % self.user_id
+        url = "%s/users/%s/vehicles?primaryOnly=true" % (IF9_BASE_URL, self.user_id)
         return self.__open(url, headers)
 
     def get_user_info(self):
         """Get user information"""
-        return self.get(self.user_id, "https://jlp-ifoa.wirelesscar.net/if9/jlr/users", self.head)
+        return self.get(self.user_id, "%s/users" % IF9_BASE_URL, self.head)
 
     def update_user_info(self, user_info_data):
         """Update user information"""
         headers = self.head.copy()
         headers["Content-Type"] = "application/vnd.wirelesscar.ngtp.if9.User-v3+json; charset=utf-8"
-        return self.post(self.user_id, "https://jlp-ifoa.wirelesscar.net/if9/jlr/users", headers, user_info_data)
+        return self.post(self.user_id, "%s/users" % IF9_BASE_URL, headers, user_info_data)
 
     def reverse_geocode(self, lat, lon):
         """Get geocode information"""
         return self.get("en",
-                        "https://jlp-ifoa.wirelesscar.net/if9/jlr/geocode/reverse/{0:f}/{0:f}".format(lat, lon),
+                        "%s/geocode/reverse/{0:f}/{1:f}".format(lat, lon) % IF9_BASE_URL,
                         self.head)
 
 
@@ -179,19 +208,22 @@ class Vehicle(dict):
         result = self.get('attributes', headers)
         return result
 
-    def get_status(self):
+    def get_status(self, key=None):
         """Get vehicle status"""
         headers = self.connection.head.copy()
         headers["Accept"] = "application/vnd.ngtp.org.if9.healthstatus-v2+json"
         result = self.get('status', headers)
+
+        if key:
+            return {d['key']: d['value'] for d in result['vehicleStatus']}[key]
+
         return result
 
     def get_health_status(self):
         """Get vehicle health status"""
         headers = self.connection.head.copy()
         headers["Accept"] = "application/vnd.wirelesscar.ngtp.if9.ServiceStatus-v4+json"
-        headers[
-            "Content-Type"] = "application/vnd.wirelesscar.ngtp.if9.StartServiceConfiguration-v3+json; charset=utf-8"
+        headers["Content-Type"] = "application/vnd.wirelesscar.ngtp.if9.StartServiceConfiguration-v3+json; charset=utf-8"
 
         vhs_data = self._authenticate_vhs()
 
@@ -214,11 +246,43 @@ class Vehicle(dict):
         result = self.get('subscriptionpackages', self.connection.head)
         return result
 
-    def get_trips(self):
+    def get_trips(self, count=1000):
         """Get the last 1000 trips associated with vehicle"""
         headers = self.connection.head.copy()
         headers["Accept"] = "application/vnd.ngtp.org.triplist-v2+json"
-        return self.get('trips?count=1000', headers)
+        return self.get('trips?count=%d' % count, headers)
+
+    def get_guardian_mode_alarms(self):
+        """Get Guardian Mode Alarms"""
+        headers = self.connection.head.copy()
+        headers["Accept"] = "application/vnd.wirelesscar.ngtp.if9.GuardianStatus-v1+json"
+        headers["Accept-Encoding"] = "gzip,deflate"
+        return self.get('gm/alarms', headers)
+
+    def get_guardian_mode_alerts(self):
+        """Get Guardian Mode Alerts"""
+        headers = self.connection.head.copy()
+        headers["Accept"] = "application/wirelesscar.GuardianAlert-v1+json"
+        headers["Accept-Encoding"] = "gzip,deflate"
+        return self.get('gm/alerts', headers)
+
+    def get_guardian_mode_status(self):
+        """Get Guardian Mode Status"""
+        headers = self.connection.head.copy()
+        headers["Accept"] = "application/vnd.wirelesscar.ngtp.if9.GuardianStatus-v1+json"
+        return self.get('gm/status', headers)
+
+    def get_guardian_mode_settings_user(self):
+        """Get Guardian Mode User Settings"""
+        headers = self.connection.head.copy()
+        headers["Accept"] = "application/vnd.wirelesscar.ngtp.if9.GuardianUserSettings-v1+json"
+        return self.get('gm/settings/user', headers)
+
+    def get_guardian_mode_settings_system(self):
+        """Get Guardian Mode System Settings"""
+        headers = self.connection.head.copy()
+        headers["Accept"] = "application/vnd.wirelesscar.ngtp.if9.GuardianSystemSettings-v1+json"
+        return self.get('gm/settings/system', headers)
 
     def get_trip(self, trip_id):
         """Get info on a specific trip"""
@@ -228,6 +292,22 @@ class Vehicle(dict):
         """Get current vehicle position"""
         return self.get('position', self.connection.head)
 
+    def get_service_status(self, service_id):
+        """Get service status"""
+        headers = self.connection.head.copy()
+        headers["Accept"] = "application/vnd.wirelesscar.ngtp.if9.ServiceStatus-v4+json"
+        return self.get('services/%s' % service_id, headers)
+
+    def get_services(self):
+        """Get active services"""
+        headers = self.connection.head.copy()
+        return self.get("services", headers)
+
+    def get_rcc_target_value(self):
+        """Get Remote Climate Target Value"""
+        headers = self.connection.head.copy()
+        return self.get('settings/ClimateControlRccTargetTemp', headers)
+
     def set_attributes(self, nickname, registration_number):
         """Set vehicle nickname and registration number"""
         attributes_data = {"nickname": nickname,
@@ -268,12 +348,39 @@ class Vehicle(dict):
         hblf_data = self.authenticate_hblf()
         return self.post("honkBlink", headers, hblf_data)
 
+    def remote_engine_start(self, pin, target_value):
+        """Start Remote Engine preconditioning"""
+        headers = self.connection.head.copy()
+        headers["Content-Type"] = "application/vnd.wirelesscar.ngtp.if9.StartServiceConfiguration-v2+json"
+        self.set_rcc_target_value(pin, target_value)
+        reon_data = self.authenticate_reon(pin)
+
+        return self.post("engineOn", headers, reon_data)
+
+    def remote_engine_stop(self, pin):
+        """Stop Remote Engine preconditioning"""
+        headers = self.connection.head.copy()
+        headers["Content-Type"] = "application/vnd.wirelesscar.ngtp.if9.StartServiceConfiguration-v2+json"
+        reoff_data = self.authenticate_reoff(pin)
+
+        return self.post("engineOff", headers, reoff_data)
+
+    def set_rcc_target_value(self, pin, target_value):
+        """Set Remote Climate Target Value (value between 31-57, 31 is LO 57 is HOT)"""
+        headers = self.connection.head.copy()
+        self.enable_provisioning_mode(pin)
+        service_parameters = {"key": "ClimateControlRccTargetTemp",
+                               "value": "%s" % str(target_value),
+                               "applied": 1}
+        self.post("settings", headers, service_parameters)
+
     def preconditioning_start(self, target_temp):
         """Start pre-conditioning for specified temperature (celsius)"""
         service_parameters = [{"key": "PRECONDITIONING",
                                "value": "START"},
                               {"key": "TARGET_TEMPERATURE_CELSIUS",
                                "value": "%s" % target_temp}]
+
         return self._preconditioning_control(service_parameters)
 
     def preconditioning_stop(self):
@@ -292,7 +399,7 @@ class Vehicle(dict):
         """Control the climate preconditioning"""
         headers = self.connection.head.copy()
         headers["Accept"] = "application/vnd.wirelesscar.ngtp.if9.ServiceStatus-v5+json"
-        headers["Content-Type"] = "application/vnd.wirelesscar.ngtp.if9.PhevService-v1+json; charset=utf"
+        headers["Content-Type"] = "application/vnd.wirelesscar.ngtp.if9.PhevService-v1+json; charset=utf-8"
 
         ecc_data = self.authenticate_ecc()
         ecc_data['serviceParameters'] = service_parameters
@@ -391,45 +498,49 @@ class Vehicle(dict):
         swu_data = self.authenticate_swu()
         swu_data["serviceCommand"] = "START"
         swu_data["startTime"] = wakeup_time
-        self._swu(swu_data)
+        return self._swu(swu_data)
 
     def delete_wakeup_time(self):
         """Stop the wakeup time"""
         swu_data = self.authenticate_swu()
         swu_data["serviceCommand"] = "END"
-        self._swu(swu_data)
+        return self._swu(swu_data)
 
     def _swu(self, swu_data):
         """Set the wakeup time for the specified time (epoch milliseconds)"""
         headers = self.connection.head.copy()
         headers["Accept"] = "application/vnd.wirelesscar.ngtp.if9.ServiceStatus-v3+json"
         headers["Content-Type"] = "application/vnd.wirelesscar.ngtp.if9.StartServiceConfiguration-v3+json; charset=utf-8"
-
         return self.post("swu", headers, swu_data)
 
+    def enable_provisioning_mode(self, pin):
+        """Enable provisioning mode """
+        self._prov_command(pin, None, "provisioning")
+
     def enable_service_mode(self, pin, expiration_time):
         """Enable service mode. Will disable at the specified time (epoch millis)"""
-        self._prov_command(pin, expiration_time, "protectionStrategy_serviceMode")
+        return self._prov_command(pin, expiration_time, "protectionStrategy_serviceMode")
+
+    def enable_guardian_mode(self, pin, expiration_time):
+        """Enable Guardian Mode until the specified time (epoch millis)"""
+        return self._gm_command(pin, expiration_time, "ACTIVE")
 
     def enable_transport_mode(self, pin, expiration_time):
         """Enable transport mode. Will be disabled at the specified time (epoch millis)"""
-        self._prov_command(pin, expiration_time, "protectionStrategy_transportMode")
+        return self._prov_command(pin, expiration_time, "protectionStrategy_transportMode")
 
     def enable_privacy_mode(self, pin):
         """Enable privacy mode. Will disable journey logging"""
-        self._prov_command(pin, None, "privacySwitch_on")
+        return self._prov_command(pin, None, "privacySwitch_on")
 
     def disable_privacy_mode(self, pin):
         """Disable privacy mode. Will enable journey logging"""
-        self._prov_command(pin, None, "privacySwitch_off")
+        return self._prov_command(pin, None, "privacySwitch_off")
 
     def _prov_command(self, pin, expiration_time, mode):
         """Send prov endpoint commands. Used for service/transport/privacy mode"""
         headers = self.connection.head.copy()
-        headers["Accept"] = "application/vnd.wirelesscar.ngtp.if9.ServiceStatus-v4+json"
-        headers[
-            "Content-Type"] = "application/vnd.wirelesscar.ngtp.if9.StartServiceConfiguration-v3+json; charset=utf-8"
-
+        headers["Content-Type"] = "application/vnd.wirelesscar.ngtp.if9.StartServiceConfiguration-v3+json"
         prov_data = self.authenticate_prov(pin)
 
         prov_data["serviceCommand"] = mode
@@ -438,6 +549,16 @@ class Vehicle(dict):
 
         return self.post("prov", headers, prov_data)
 
+    def _gm_command(self, pin, expiration_time, status):
+        """Send GM toggle command"""
+        headers = self.connection.head.copy()
+        headers["Accept"] = "application/vnd.wirelesscar.ngtp.if9.GuardianAlarmList-v1+json"
+        gm_data = self.authenticate_gm(pin)
+        gm_data["endTime"] = expiration_time
+        gm_data["status"] = status
+
+        return self.post("gm/alarms", headers, gm_data)
+
     def _authenticate_vhs(self):
         """Authenticate to vhs and get token"""
         return self._authenticate_empty_pin_protected_service("VHS")
@@ -489,10 +610,21 @@ class Vehicle(dict):
         """Authenticate to aloff"""
         return self._authenticate_pin_protected_service(pin, "ALOFF")
 
+    def authenticate_reon(self, pin):
+        """Authenticate to reon"""
+        return self._authenticate_pin_protected_service(pin, "REON")
+
+    def authenticate_reoff(self, pin):
+        """Authenticate to reoff"""
+        return self._authenticate_pin_protected_service(pin, "REOFF")
+
     def authenticate_prov(self, pin):
         """Authenticate to PROV service"""
         return self._authenticate_pin_protected_service(pin, "PROV")
 
+    def authenticate_gm(self, pin):
+        return self._authenticate_pin_protected_service(pin, "GM")
+
     def _authenticate_pin_protected_service(self, pin, service_name):
         """Authenticate to specified service with the provided PIN"""
         data = {
@@ -505,9 +637,9 @@ class Vehicle(dict):
 
     def post(self, command, headers, data):
         """Utility command to post data to VHS"""
-        return self.connection.post(command, 'https://jlp-ifoa.wirelesscar.net/if9/jlr/vehicles/%s' % self.vin,
+        return self.connection.post(command, '%s/vehicles/%s' % (IF9_BASE_URL, self.vin),
                                     headers, data)
 
     def get(self, command, headers):
         """Utility command to get vehicle data from API"""
-        return self.connection.get(command, 'https://jlp-ifoa.wirelesscar.net/if9/jlr/vehicles/%s' % self.vin, headers)
+        return self.connection.get(command, '%s/vehicles/%s' % (IF9_BASE_URL, self.vin), headers)
